/*
 * otonat: simple and fast 1:1 NAT for IPv4
 * 
 * written by Manuel Kasper <mkasper@monzoon.net>
 * (c) 2011 Monzoon Networks AG. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *   3. Neither the name of Monzoon Networks AG nor the
 *      names of its contributors may be used to endorse or promote products
 *      derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MONZOON NETWORKS AG BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: otonat.c 3849 2015-11-10 11:55:19Z monzoon\mkasper $
 */

#include "../otonat.h"

#include <sys/types.h>
#include <sys/param.h>
#include <sys/conf.h>
#include <sys/stat.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/module.h>
#include <sys/sysctl.h>
#include <sys/kernel.h>
#include <sys/socket.h>
#include <net/pfil.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <sys/queue.h>
#include <net/if.h>
#include <net/if_var.h>
#include <net/if_types.h>
#include <sys/lock.h>
#include <sys/endian.h>

#if __FreeBSD_version > 700055
#include <sys/rmlock.h>
#define HAVE_RMLOCK 1
#else
#include <sys/mutex.h>
#endif

#define OTONAT_HASH_SIZE	1024

#define OTONAT_HASHKEY(addr) (be32toh(addr) % OTONAT_HASH_SIZE)

static int otonat_handler(module_t mod, int cmd, void *arg);
static int otonat_load(void);
static int otonat_unload(void);
static int otonat_hook_pfil(void *arg, struct mbuf **mp, struct ifnet *ifp, int dir, struct inpcb *inp);
static int otonat_ioctl(struct cdev *dev, u_long cmd, caddr_t addr, int flags, struct thread *td);

static moduledata_t otonat_mod = {
	"otonat",
	otonat_handler,
	0
};

struct otonat_mapping {
	struct ifnet	*ifp;
	struct ifnet	*ifp2;
	in_addr_t		internal_addr;
	in_addr_t		external_addr;
};

struct otonat_hash_entry {
	struct otonat_mapping			*mapping;
	SLIST_ENTRY(otonat_hash_entry)	entries;
};

struct otonat_excl_entry {
	in_addr_t						addr;
	SLIST_ENTRY(otonat_excl_entry)	entries;
};

static struct cdevsw otonat_cdevsw = {
	.d_version = D_VERSION,
	.d_ioctl = otonat_ioctl,
	.d_name = "ioctl"
};

static int pfil_count_in = 0;
static int pfil_count_out = 0;
static SLIST_HEAD(,otonat_hash_entry) *hash_int2ext;
static SLIST_HEAD(,otonat_hash_entry) *hash_ext2int;
static SLIST_HEAD(,otonat_excl_entry) *hash_excl;
static struct cdev *ctldev;

#if HAVE_RMLOCK
static struct rmlock natmap_lock;
#define NATMAP_RLOCK()		struct rm_priotracker tracker; \
							rm_rlock(&natmap_lock, &tracker);
#define NATMAP_RUNLOCK()	rm_runlock(&natmap_lock, &tracker);
#define NATMAP_WLOCK()		rm_wlock(&natmap_lock);
#define NATMAP_WUNLOCK()	rm_wunlock(&natmap_lock);
#else
static struct mtx natmap_mtx;
#define NATMAP_RLOCK()		mtx_lock(&natmap_mtx);
#define NATMAP_RUNLOCK()	mtx_unlock(&natmap_mtx);
#define NATMAP_WLOCK()		mtx_lock(&natmap_mtx);
#define NATMAP_WUNLOCK()	mtx_unlock(&natmap_mtx);
#endif

SYSCTL_DECL(_net_inet);
SYSCTL_DECL(_net_inet_ip);
SYSCTL_NODE(_net_inet_ip, OID_AUTO, otonat, CTLFLAG_RW, 0, "otonat");
SYSCTL_UINT(_net_inet_ip_otonat, OID_AUTO, pfil_count_in, CTLFLAG_RD, &pfil_count_in, 0, "pfil in packet count");
SYSCTL_UINT(_net_inet_ip_otonat, OID_AUTO, pfil_count_out, CTLFLAG_RD, &pfil_count_out, 0, "pfil out packet count");

MALLOC_DEFINE(M_OTONAT, "otonat", "1:1 NAT module");

static struct otonat_mapping* otonat_lookup_external_addr(struct ifnet *ifp, in_addr_t addr);
static struct otonat_mapping* otonat_lookup_internal_addr(struct ifnet *ifp, in_addr_t addr);
static int otonat_insert_mapping(struct otonat_mapping *mapping);
static void otonat_remove_mapping(struct otonat_mapping *mapping);
static void otonat_flush_mappings(void);
static int otonat_insert_exclusion(in_addr_t addr);
static int otonat_remove_exclusion(in_addr_t addr);
static int otonat_lookup_exclusion(in_addr_t addr);
static void otonat_flush_exclusions(void);
static uint16_t adjust_checksum(uint16_t csum, in_addr_t oldaddr, in_addr_t newaddr, int isudp);

static int otonat_handler(module_t mod, int cmd, void *arg) {
	int error = 0;
	
	switch (cmd) {
		case MOD_LOAD:
			error = otonat_load();
			break;
		case MOD_UNLOAD:
			error = otonat_unload();
			break;
		default:
			error = EOPNOTSUPP;
	}
	
	return (error);
}

static int otonat_load(void) {
	struct pfil_head *pfh_inet;
	int i;
	
#if HAVE_RMLOCK
	rm_init(&natmap_lock, "otonat map");
#else
	mtx_init(&natmap_mtx, "otonat map", NULL, MTX_DEF);
#endif

	/* Allocate memory for simple hash tables that help with finding
	   mappings from internal to external addresses and vice versa.
	   The least significant bits of the address are used as a hash key */
	hash_int2ext = malloc(OTONAT_HASH_SIZE * sizeof(*hash_int2ext), M_OTONAT, M_NOWAIT);
	hash_ext2int = malloc(OTONAT_HASH_SIZE * sizeof(*hash_ext2int), M_OTONAT, M_NOWAIT);
	
	/* Allocate memory for exclusions hash */
	hash_excl = malloc(OTONAT_HASH_SIZE * sizeof(*hash_excl), M_OTONAT, M_NOWAIT);
	
	for (i = 0; i < OTONAT_HASH_SIZE; i++) {
		SLIST_INIT(&hash_int2ext[i]);
		SLIST_INIT(&hash_ext2int[i]);
		SLIST_INIT(&hash_excl[i]);
	}
	
	/* hook into pfil */
	pfh_inet = pfil_head_get(PFIL_TYPE_AF, AF_INET);
	if (pfh_inet == NULL) {
		printf("otonat: cannot get pfil head\n");
		return ENOPROTOOPT;
	}
		
	pfil_add_hook(otonat_hook_pfil, NULL, PFIL_IN | PFIL_OUT | PFIL_WAITOK, pfh_inet);
	
	ctldev = make_dev(&otonat_cdevsw, 0, 0, 0, S_IRUSR | S_IWUSR, "otonat");
	if (ctldev == 0) {
		printf("otonat: cannot install control device");
		return EIO;
	}
	
	printf("otonat " OTONAT_VERSION " loaded\n");
	
	return 0;
}

static int otonat_unload(void) {
	struct pfil_head *pfh_inet;
	
	/* remove control device */
	destroy_dev(ctldev);
	
	/* unhook from pfil */
	pfh_inet = pfil_head_get(PFIL_TYPE_AF, AF_INET);
	if (pfh_inet == NULL) {
		printf("otonat: cannot get pfil head\n");
		return ENOPROTOOPT;
	}
		
	pfil_remove_hook(otonat_hook_pfil, NULL, PFIL_IN | PFIL_OUT | PFIL_WAITOK, pfh_inet);
	
	/* free all NAT mappings and exclusions */
	otonat_flush_mappings();
	otonat_flush_exclusions();
	
	/* free hash tables */
	free(hash_int2ext, M_OTONAT);
	free(hash_ext2int, M_OTONAT);
	free(hash_excl, M_OTONAT);
	
#if HAVE_RMLOCK
	rm_destroy(&natmap_lock);
#else
	mtx_destroy(&natmap_mtx);
#endif
	
	printf("otonat v" OTONAT_VERSION " unloaded\n");
	
	return 0;
}

static int otonat_hook_pfil(void *arg, struct mbuf **mp, struct ifnet *ifp, int dir, struct inpcb *inp) {
	struct ip *iph;
	struct otonat_mapping *mapping;
	in_addr_t oldaddr = 0, newaddr = 0;
	
	if ((*mp)->m_pkthdr.len < (int)sizeof(struct ip))
		return 0;
	
	iph = mtod(*mp, struct ip *);
	
	/* Only deal with IPv4 packets */
	if (iph->ip_v != 4) {
		printf("otonat: got non-v4 packet\n");
		return 0;
	}
	
	NATMAP_RLOCK();
	
	if (dir == PFIL_IN) {
		/* search for packet's destination address in external_addr field */
		mapping = otonat_lookup_external_addr(ifp, iph->ip_dst.s_addr);
		if (mapping != 0) {
			if (!otonat_lookup_exclusion(iph->ip_src.s_addr)) {			
				oldaddr = iph->ip_dst.s_addr;
				newaddr = mapping->internal_addr;
				iph->ip_dst.s_addr = mapping->internal_addr;
			}
		}
		
		pfil_count_in++;
	} else if (dir == PFIL_OUT) {		
		/* search for packet's source address in internal_addr field */
		mapping = otonat_lookup_internal_addr(ifp, iph->ip_src.s_addr);
		if (mapping != 0) {
			if (!otonat_lookup_exclusion(iph->ip_dst.s_addr)) {
				oldaddr = iph->ip_src.s_addr;
				newaddr = mapping->external_addr;
				iph->ip_src.s_addr = mapping->external_addr;
			}
		}
		
		pfil_count_out++;
	}
	
	NATMAP_RUNLOCK();
	
	if (oldaddr != newaddr) {
		/* the address has changed - we may have to adjust the inner
		   header's checksum for some protocols, but only for the
		   first fragment */
		
#if (__FreeBSD_version >= 1000019)
		if ((iph->ip_off & htons(IP_OFFMASK)) == 0) {
#else
		if ((iph->ip_off & IP_OFFMASK) == 0) {
#endif
			switch (iph->ip_p) {
				case IPPROTO_UDP: {
					struct udphdr *udp;
		
					*mp = m_pullup(*mp, sizeof(struct ip) + sizeof(struct udphdr));
					if (*mp != NULL) {
						udp = (struct udphdr*)(mtod(*mp, char *) + sizeof(struct ip));
						udp->uh_sum = adjust_checksum(udp->uh_sum, oldaddr, newaddr, 1);
					}
					break;
				}
				case IPPROTO_TCP: {	
					struct tcphdr *tcp;
		
					*mp = m_pullup(*mp, sizeof(struct ip) + sizeof(struct tcphdr));
					if (*mp != NULL) {
						tcp = (struct tcphdr*)(mtod(*mp, char *) + sizeof(struct ip));
						tcp->th_sum = adjust_checksum(tcp->th_sum, oldaddr, newaddr, 0);
					}
					break;
				}
			}
		}
	}
	
	return 0;
}

static int otonat_ioctl(struct cdev *dev, u_long cmd, caddr_t arg, int flags, struct thread *td) {
	int error = 0;

	switch (cmd) {
	
	case OTOIOCADDMAPPING: {
		struct otonat_ioc_mapping *iocmapping = (struct otonat_ioc_mapping*)arg;
		struct ifnet *ifp = NULL;
		struct ifnet *ifp2 = NULL;
		struct otonat_mapping *mapping;
		char tmpifn[IF_NAMESIZE*2+1];
		char *tmpifnp = tmpifn;
		char *token;
		
		strncpy(tmpifn, iocmapping->if_name, sizeof(tmpifn));
		while ((token = strsep(&tmpifnp, ",")) != NULL) {
			if (ifp == NULL) {
				ifp = ifunit(token);
				if (ifp == NULL) {
					error = EINVAL;
					break;
				}
			} else {
				ifp2 = ifunit(token);
				if (ifp2 == NULL) {
					error = EINVAL;
					break;
				}
			}
		}
		
		mapping = malloc(sizeof(struct otonat_mapping), M_OTONAT, M_NOWAIT);
		mapping->ifp = ifp;
		mapping->ifp2 = ifp2;
		mapping->internal_addr = iocmapping->internal_addr;
		mapping->external_addr = iocmapping->external_addr;
		
		NATMAP_WLOCK();
		error = otonat_insert_mapping(mapping);
		if (error != 0)
			free(mapping, M_OTONAT);
		NATMAP_WUNLOCK();
		
		break;
	}
	
	case OTOIOCDELMAPPING: {
		struct otonat_ioc_mapping *iocmapping = (struct otonat_ioc_mapping*)arg;
		struct ifnet *ifp = NULL;
		struct ifnet *ifp2 = NULL;
		struct otonat_mapping *mapping;
		char tmpifn[IF_NAMESIZE*2+1];
		char *tmpifnp = tmpifn;
		char *token;
		
		strncpy(tmpifn, iocmapping->if_name, sizeof(tmpifn));
		while ((token = strsep(&tmpifnp, ",")) != NULL) {
			if (ifp == NULL) {
				ifp = ifunit(token);
				if (ifp == NULL) {
					error = EINVAL;
					break;
				}
			} else {
				ifp2 = ifunit(token);
				if (ifp2 == NULL) {
					error = EINVAL;
					break;
				}
			}
		}
		
		NATMAP_WLOCK();
		
		/* The user may specify either the internal or the external address when deleting rules */
		if (iocmapping->external_addr != 0)
			mapping = otonat_lookup_external_addr(ifp, iocmapping->external_addr);
		else
			mapping = otonat_lookup_internal_addr(ifp, iocmapping->internal_addr);
		
		if (mapping == NULL) {
			error = ENOENT;
			NATMAP_WUNLOCK();
			break;
		}
		
		otonat_remove_mapping(mapping);
		free(mapping, M_OTONAT);
		
		NATMAP_WUNLOCK();
		
		break;
	}
	
	case OTOIOCGETMAPPINGS: {
		int i;
		struct otonat_hash_entry *ent;
		struct otonat_ioc_getreq *req = (struct otonat_ioc_getreq*)arg;
		struct otonat_ioc_mapping res;
		
		NATMAP_RLOCK();
		
		req->n = 0;
		
		/* copy all entries out to userland */
		for (i = 0; i < OTONAT_HASH_SIZE; i++) {
			SLIST_FOREACH(ent, &hash_int2ext[i], entries) {
				char tmpifn[IF_NAMESIZE*2+1];
				
				if (req->n >= req->bufsize) {
					/* we've run out of space - bail out */
					error = ENOSPC;
					NATMAP_RUNLOCK();
					break;
				}
				
				strcpy(tmpifn, ent->mapping->ifp->if_xname);
				if (ent->mapping->ifp2 != NULL) {
					strcat(tmpifn, ",");
					strcat(tmpifn, ent->mapping->ifp2->if_xname);
				}
				
				copystr(tmpifn, res.if_name, sizeof(res.if_name), NULL);
				res.internal_addr = ent->mapping->internal_addr;
				res.external_addr = ent->mapping->external_addr;
				copyout(&res, &req->u_buf.mapping_buf[req->n], sizeof(struct otonat_ioc_mapping));
				
				req->n++;
			}
		}
		
		NATMAP_RUNLOCK();
		
		break;
	}
	
	case OTOIOCFLUSHMAPPINGS: {
		NATMAP_WLOCK();
		otonat_flush_mappings();
		NATMAP_WUNLOCK();
		break;
	}
	
	case OTOIOCGETVERSION:
		strcpy((char*)arg, OTONAT_VERSION);
		break;
	
	case OTOIOCADDEXCLUSION: {
		NATMAP_WLOCK();
		error = otonat_insert_exclusion(*((in_addr_t*)arg));
		NATMAP_WUNLOCK();
		break;
	}
	
	case OTOIOCDELEXCLUSION: {
		NATMAP_WLOCK();
		error = otonat_remove_exclusion(*((in_addr_t*)arg));
		NATMAP_WUNLOCK();
		break;
	}
	
	case OTOIOCGETEXCLUSIONS: {
		int i;
		struct otonat_excl_entry *ent;
		struct otonat_ioc_getreq *req = (struct otonat_ioc_getreq*)arg;
		
		NATMAP_RLOCK();
		
		req->n = 0;
		
		/* copy all entries out to userland */
		for (i = 0; i < OTONAT_HASH_SIZE; i++) {
			SLIST_FOREACH(ent, &hash_excl[i], entries) {
				if (req->n >= req->bufsize) {
					/* we've run out of space - bail out */
					error = ENOSPC;
					NATMAP_RUNLOCK();
					break;
				}
				
				copyout(&ent->addr, &req->u_buf.addr_buf[req->n], sizeof(in_addr_t));
				
				req->n++;
			}
		}
		
		NATMAP_RUNLOCK();
		
		break;
	}
	
	case OTOIOCFLUSHEXCLUSIONS: {
		NATMAP_WLOCK();
		otonat_flush_exclusions();
		NATMAP_WUNLOCK();
		break;
	}
	
	default:
		error = ENODEV;
		break;
	}
	
	return (error);
}

/* Look up a mapping given the external address. Assumes that natmap_lock has
   already been aquired. Returns 0 if not found. */
static struct otonat_mapping* otonat_lookup_external_addr(struct ifnet *ifp, in_addr_t addr) {
	struct otonat_hash_entry *ent;
	int hashkey = OTONAT_HASHKEY(addr);
	
	SLIST_FOREACH(ent, &hash_ext2int[hashkey], entries) {
		if ((ent->mapping->ifp == ifp || ent->mapping->ifp2 == ifp) && ent->mapping->external_addr == addr)
			return ent->mapping;
	}
	
	return 0;
}

/* Look up a mapping given the internal address. Assumes that natmap_lock has
   already been aquired. Returns 0 if not found. */
static struct otonat_mapping* otonat_lookup_internal_addr(struct ifnet *ifp, in_addr_t addr) {
	struct otonat_hash_entry *ent;
	int hashkey = OTONAT_HASHKEY(addr);
	
	SLIST_FOREACH(ent, &hash_int2ext[hashkey], entries) {
		if ((ent->mapping->ifp == ifp || ent->mapping->ifp2 == ifp) && ent->mapping->internal_addr == addr)
			return ent->mapping;
	}
	
	return 0;
}

/* Insert a new mapping. The object pointed to by mapping is not copied. The caller must
   have a writer lock on natmap_lock. */
static int otonat_insert_mapping(struct otonat_mapping *mapping) {
	int hashkey_int, hashkey_ext;
	struct otonat_hash_entry *hent_int, *hent_ext;
	
	/* check the new mapping is unique */
	if (otonat_lookup_internal_addr(mapping->ifp, mapping->internal_addr) != 0 || 
		otonat_lookup_external_addr(mapping->ifp, mapping->external_addr) != 0)
		return EEXIST;
	
	/* calculate hash keys */
	hashkey_int = OTONAT_HASHKEY(mapping->internal_addr);
	hashkey_ext = OTONAT_HASHKEY(mapping->external_addr);
	
	/* allocate hash entries */
	hent_int = malloc(sizeof(struct otonat_hash_entry), M_OTONAT, M_NOWAIT);
	if (hent_int == 0)
		return ENOMEM;
	
	hent_int->mapping = mapping;
	
	hent_ext = malloc(sizeof(struct otonat_hash_entry), M_OTONAT, M_NOWAIT);
	if (hent_ext == 0) {
		free(hent_int, M_OTONAT);
		return ENOMEM;
	}
	hent_ext->mapping = mapping;
	
	/* and insert into the hash tables */
	SLIST_INSERT_HEAD(&hash_int2ext[hashkey_int], hent_int, entries);
	SLIST_INSERT_HEAD(&hash_ext2int[hashkey_ext], hent_ext, entries);
	
	return 0;
}

/* Remove a mapping. The mapping object itself is not freed. The caller must
   have a writer lock on natmap_lock. */
static void otonat_remove_mapping(struct otonat_mapping *mapping) {
	struct otonat_hash_entry *ent;
	int hashkey_int, hashkey_ext;
	
	/* calculate hash keys */
	hashkey_int = OTONAT_HASHKEY(mapping->internal_addr);
	hashkey_ext = OTONAT_HASHKEY(mapping->external_addr);
	
	SLIST_FOREACH(ent, &hash_int2ext[hashkey_int], entries) {
		if (ent->mapping == mapping) {
			SLIST_REMOVE(&hash_int2ext[hashkey_int], ent, otonat_hash_entry, entries);
			free(ent, M_OTONAT);
			break;
		}
	}
	
	SLIST_FOREACH(ent, &hash_ext2int[hashkey_ext], entries) {
		if (ent->mapping == mapping) {
			SLIST_REMOVE(&hash_ext2int[hashkey_ext], ent, otonat_hash_entry, entries);
			free(ent, M_OTONAT);
			break;
		}
	}
}

/* Remove and free all mappings. The caller must have a writer lock on natmap_lock. */
static void otonat_flush_mappings(void) {
	int i;
	for (i = 0; i < OTONAT_HASH_SIZE; i++) {
		while (!SLIST_EMPTY(&hash_int2ext[i])) {
			struct otonat_hash_entry *ent = SLIST_FIRST(&hash_int2ext[i]);
			SLIST_REMOVE_HEAD(&hash_int2ext[i], entries);
			free(ent->mapping, M_OTONAT);
			free(ent, M_OTONAT);
		}
		while (!SLIST_EMPTY(&hash_ext2int[i])) {
			struct otonat_hash_entry *ent = SLIST_FIRST(&hash_ext2int[i]);
			SLIST_REMOVE_HEAD(&hash_ext2int[i], entries);
			free(ent, M_OTONAT);
		}
	}
	printf("otonat: mappings flushed\n");
}

/* Insert a new exclusion. The caller must have a writer lock on natmap_lock. */
static int otonat_insert_exclusion(in_addr_t addr) {
	struct otonat_excl_entry *hent;
	
	/* check that the new exclusion is unique */
	if (otonat_lookup_exclusion(addr))
		return EEXIST;
	
	/* allocate hash entries */
	hent = malloc(sizeof(struct otonat_excl_entry), M_OTONAT, M_NOWAIT);
	if (hent == 0)
		return ENOMEM;
	
	hent->addr = addr;
	
	/* and insert into the hash tables */
	SLIST_INSERT_HEAD(&hash_excl[OTONAT_HASHKEY(addr)], hent, entries);
	
	return 0;
}

/* Remove an exclusion. The caller must have a writer lock on natmap_lock. */
static int otonat_remove_exclusion(in_addr_t addr) {
	struct otonat_excl_entry *ent;
	int hashkey;
	
	hashkey = OTONAT_HASHKEY(addr);
	
	SLIST_FOREACH(ent, &hash_excl[hashkey], entries) {
		if (ent->addr == addr) {
			SLIST_REMOVE(&hash_excl[hashkey], ent, otonat_excl_entry, entries);
			free(ent, M_OTONAT);
			return 0;
		}
	}
	return ENOENT;
}

static int otonat_lookup_exclusion(in_addr_t addr) {
	struct otonat_excl_entry *ent;
	int hashkey = OTONAT_HASHKEY(addr);
	
	SLIST_FOREACH(ent, &hash_excl[hashkey], entries) {
		if (ent->addr == addr) {
			return 1;
		}
	}
	
	return 0;
}

/* Remove and free all exclusions. The caller must have a writer lock on natmap_lock. */
static void otonat_flush_exclusions(void) {
	int i;
	for (i = 0; i < OTONAT_HASH_SIZE; i++) {
		while (!SLIST_EMPTY(&hash_excl[i])) {
			struct otonat_excl_entry *ent = SLIST_FIRST(&hash_excl[i]);
			SLIST_REMOVE_HEAD(&hash_excl[i], entries);
			free(ent, M_OTONAT);
		}
	}
	printf("otonat: exclusions flushed\n");
}

static uint16_t adjust_checksum(uint16_t csum, in_addr_t oldaddr, in_addr_t newaddr, int isudp) {
	uint32_t x = csum;
	
	if (isudp && csum == 0)	/* as per RFC 768, 0 means no checksum was generated */
		return 0;
	
	/* add difference between upper/lower two address bytes */
	x += ((oldaddr >> 16) & 0xFFFF) - ((newaddr >> 16) & 0xFFFF);
	x += (oldaddr & 0xFFFF) - (newaddr & 0xFFFF);
	x = (x >> 16) + (x & 0xFFFF);
	x &= 0xFFFF;
	
	if (x == 0)
		x = 0xFFFF;
	
	return x;
}

DECLARE_MODULE(otonat, otonat_mod, SI_SUB_EXEC, SI_ORDER_ANY);
