/*
 * otonat: simple and fast 1:1 NAT for IPv4
 * 
 * written by Manuel Kasper <mkasper@monzoon.net>
 * (c) 2011 Monzoon Networks AG. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *   3. Neither the name of Monzoon Networks AG nor the
 *      names of its contributors may be used to endorse or promote products
 *      derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MONZOON NETWORKS AG BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: otonatctl.c 2062 2011-07-11 20:24:29Z monzoon\mkasper $
 */

#include <stdio.h>
#include <err.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sysexits.h>

#include "../otonat.h"

enum {
	CMD_ADD = 1,
	CMD_DELETE,
	CMD_FLUSH,
	CMD_LIST
};

enum {
	MODE_MAPPINGS = 1,
	MODE_EXCLUSIONS
};

int fd;

void add_mapping(char *interface, char *internal_addr, char *external_addr);
void delete_mapping(char *interface, char *internal_addr, char *external_addr);
void list_mappings(void);
void flush_mappings(void);

void add_exclusion(char *excl_addr);
void delete_exclusion(char *excl_addr);
void list_exclusions(void);
void flush_exclusions(void);

void usage(void);
void check_version(void);

int main(int argc, char *argv[]) {
	
	int ch;
	int cmd = 0;
	int mode = 0;
	char *internal_addr = 0;
	char *external_addr = 0;
	char *interface = 0;

	while ((ch = getopt(argc, argv, "ade:fi:lmn:xh")) != -1) {
		switch (ch) {
		case 'a':
			cmd = CMD_ADD;
			break;
		case 'd':
			cmd = CMD_DELETE;
			break;
		case 'e':
			external_addr = optarg;
			break;
		case 'f':
			cmd = CMD_FLUSH;
			break;
		case 'i':
			internal_addr = optarg;
			break;
		case 'l':
			cmd = CMD_LIST;
			break;
		case 'm':
			mode = MODE_MAPPINGS;
			break;
		case 'n':
			interface = optarg;
			break;
		case 'x':
			mode = MODE_EXCLUSIONS;
			break;
		case 'h':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	
	/* check if we have enough arguments */
	if (mode == MODE_MAPPINGS) {
		if (cmd == CMD_ADD && !(interface && internal_addr && external_addr))
			usage();
		else if (cmd == CMD_DELETE && !(interface && (internal_addr || external_addr)))
			usage();
	} else if (mode == MODE_EXCLUSIONS) {
		if ((cmd == CMD_ADD || cmd == CMD_DELETE) && !external_addr)
			usage();
	} else {
		usage();
	}
	
	fd = open("/dev/otonat", O_RDWR);
	if (fd == -1)
		err(EX_NOINPUT, "open");
	
	check_version();
	
	if (mode == MODE_MAPPINGS) {
		switch (cmd) {
			case CMD_ADD:
				add_mapping(interface, internal_addr, external_addr);
				break;
			case CMD_DELETE:
				delete_mapping(interface, internal_addr, external_addr);
				break;
			case CMD_LIST:
				list_mappings();
				break;
			case CMD_FLUSH:
				flush_mappings();
				break;
			default:
				usage();
		}
	} else if (mode == MODE_EXCLUSIONS) {
		switch (cmd) {
			case CMD_ADD:
				add_exclusion(external_addr);
				break;
			case CMD_DELETE:
				delete_exclusion(external_addr);
				break;
			case CMD_LIST:
				list_exclusions();
				break;
			case CMD_FLUSH:
				flush_exclusions();
				break;
			default:
				usage();
		}
	}
	
	close(fd);
	
	return 0;
}

void add_mapping(char *interface, char *internal_addr, char *external_addr) {
	struct otonat_ioc_mapping map;
	strcpy(map.if_name, interface);
	map.internal_addr = inet_addr(internal_addr);
	map.external_addr = inet_addr(external_addr);
	if (ioctl(fd, OTOIOCADDMAPPING, &map) != 0) {
		warn("ioctl");
	}
}

void delete_mapping(char *interface, char *internal_addr, char *external_addr) {
	struct otonat_ioc_mapping map;
	strcpy(map.if_name, interface);
	map.internal_addr = (internal_addr != 0 ? inet_addr(internal_addr) : 0);
	map.external_addr = (external_addr != 0 ? inet_addr(external_addr) : 0);
	if (ioctl(fd, OTOIOCDELMAPPING, &map) != 0) {
		warn("ioctl");
	}
}

void list_mappings(void) {
	struct otonat_ioc_getreq req;
	int i;
	
	req.bufsize = 1000;
	req.u_buf.mapping_buf = malloc(1000 * sizeof(struct otonat_ioc_mapping));
	if (ioctl(fd, OTOIOCGETMAPPINGS, &req) != 0) {
		warn("ioctl");
	}
	printf("%d mappings:\n", req.n);
	
	for (i = 0; i < req.n; i++) {
		char intaddr[INET_ADDRSTRLEN], extaddr[INET_ADDRSTRLEN];
		inet_ntop(AF_INET, &req.u_buf.mapping_buf[i].internal_addr, intaddr, INET_ADDRSTRLEN);
		inet_ntop(AF_INET, &req.u_buf.mapping_buf[i].external_addr, extaddr, INET_ADDRSTRLEN);
		printf("%s -> %s (%s)\n", intaddr, extaddr, req.u_buf.mapping_buf[i].if_name);
	}
}

void flush_mappings(void) {
	if (ioctl(fd, OTOIOCFLUSHMAPPINGS, 0) != 0) {
		warn("ioctl");
	}
}

void add_exclusion(char *excl_addr) {
	in_addr_t addr = inet_addr(excl_addr);
	if (ioctl(fd, OTOIOCADDEXCLUSION, &addr) != 0) {
		warn("ioctl");
	}
}

void delete_exclusion(char *excl_addr) {
	in_addr_t addr = inet_addr(excl_addr);
	if (ioctl(fd, OTOIOCDELEXCLUSION, &addr) != 0) {
		warn("ioctl");
	}
}

void list_exclusions(void) {
	struct otonat_ioc_getreq req;
	int i;
	
	req.bufsize = 1000;
	req.u_buf.addr_buf = malloc(1000 * sizeof(struct otonat_ioc_mapping));
	if (ioctl(fd, OTOIOCGETEXCLUSIONS, &req) != 0) {
		warn("ioctl");
	}
	printf("%d exclusions:\n", req.n);
	
	for (i = 0; i < req.n; i++) {
		char excladdr[INET_ADDRSTRLEN];
		inet_ntop(AF_INET, &req.u_buf.addr_buf[i], excladdr, INET_ADDRSTRLEN);
		printf("%s\n", excladdr);
	}
}

void flush_exclusions(void) {
	if (ioctl(fd, OTOIOCFLUSHEXCLUSIONS, 0) != 0) {
		warn("ioctl");
	}
}

void check_version(void) {
	/* check that the kernel version of otonat matches our version */
	char kversionbuf[OTONAT_VERSION_LENGTH];
	if (ioctl(fd, OTOIOCGETVERSION, kversionbuf) != 0) {
		warn("ioctl");
	}
	
	if (strcmp(OTONAT_VERSION, kversionbuf) != 0) {
		errx(EX_SOFTWARE, "Version mismatch! (kernel: %s, userland: %s)", kversionbuf, OTONAT_VERSION);
	}
}

void usage(void) {
	fprintf(stderr, "otonat " OTONAT_VERSION "\n\n");
	fprintf(stderr, "usage:\n");
	fprintf(stderr, "otonatctl -a|-d|-l|-f -m|-x [-i internal_addr] [-e external_addr] [-n interface]\n\n");
	fprintf(stderr, "          -a: add a new mapping/exclusion\n");
	fprintf(stderr, "          -d: delete a mapping/exclusion\n");
	fprintf(stderr, "          -l: list all mappings/exclusions\n");
	fprintf(stderr, "          -f: flush all mappings/exclusions\n\n");
	fprintf(stderr, "          -m: act on mappings (-i/-e/-n required)\n");
	fprintf(stderr, "          -x: act on exclusions (-e required for add/delete)\n\n");
	exit(EX_USAGE);
}
